<?php

use BitFire\Config;
use BitFire\Malware;
use BitFire\Malware_List;
use BitFire\ScanConfig;
use ThreadFin\Effect;
use ThreadFin\FileData;
use ThreadFin\FileMod;

use const BitFire\WAF_ROOT;

use function ThreadFin\partial_right;
use function ThreadFin\partial;
use function BitFire\scan_filesystem;
use function BitFireSvr\get_wordpress_version;
use function BitFireSvr\hash_file3;
use function ThreadFin\dbg;
use function ThreadFin\debug;
use function ThreadFin\en_json;
use function ThreadFin\file_index;
use function ThreadFin\get_hidden_file;
use function ThreadFin\http2;
use function ThreadFin\un_json;

if (php_sapi_name() != "cli") {
    die("This is a command line script");
}

function usage($argv) {
    echo "BitFire uses machine learning to identify malware. Machine learning identifies unknown variable names\n";
    echo "and unusual character frequencies (excessive ()[];\x, for example). Malware is identified when a\n";
    echo "percentage of unknown variable names, or unusual character frequency exceeds threshold UPPER(-Q, -N).\n";
    echo "\n";
    echo "If a different threshold is exceeded lower(-q, -n) then malware will be identified if a dynamic\n";
    echo "function call (call_user_func(), \$dynamic()) is also found in the file.\n";
    echo "\n";
    echo "Set the -q and -n options to be about 1/2 the -Q and -N options. Reduce both sets of numbers until\n";
    echo "just about where you see false positives.\n\n";

    echo "usage: php " . $argv[0] . " -p /var/www/wordpress -s -i\n";
    echo " -p <path>           - (config) path to the wordpress install\n";
    echo " -s                  - scan the filesystem\n";
    echo " -i                  - interactive malware repair\n";
    echo " -a                  - output all malware info\n";
    echo " -o                  - output just file names\n";
    echo " -h                  - hide progress indication\n";
    echo " -m                  - monochrome output\n\n";
    echo " -u                  - report unknown core files (default: yes)\n";
    echo " -U                  - reporting unknown plugin/theme files (default: no)\n";
    echo " -q                  - minimum frequency to scan for dynamic functions (default: 20)\n";
    echo " -n                  - minimum percentage unknown variables to scan dynamic functions (default: 20) \n";
    echo " -Q                  - minimum frequency to mark file as malware (default: 48)\n";
    echo " -N                  - minimum percentage unknown variables to mark file as malware (default: 40)\n";
    echo " -l                  - minimum line length to scan dynamic functions (default: 768)\n";
    echo " -L                  - minimum line length to identify malware (default: 2048)\n";
    die();
}

function config_display(ScanConfig $config, string $root, bool $color) {
    $red = $green = $cyan = $normal = "";
    if ($color) {
        $red = "\e[1;31m";
        $cyan = "\e[1;36m";
        $normal = "\e[0m";
        $green = "\e[1;32m";
        $blue = "\e[1;34m";
    }

    echo "{$normal}Malware Frequency Threshold: $cyan" . $config->freq_limit . "$normal  ";
    echo "  Unknown Variable %: $cyan" . $config->random_name_per . "$normal  ";
    echo "  Max Line Length: $cyan" . $config->line_limit . "$normal\n";
    echo "Dynamic FN Freq: $cyan" . $config->fn_freq_limit . "$normal,  Dynamic FN Unknown Variable %: $cyan" . $config->fn_random_name_per . "$normal,  Dynamic FN Max Line Length: $cyan" . $config->fn_line_limit . "$normal\n";
    echo "Scanning: {$blue}$root{$normal}\n\n";
}

function min_display(Malware $malware, bool $color, ScanConfig $config) {
    $red = $green = $cyan = $normal = "";
    if ($color) {
        $red = "\e[1;31m";
        $cyan = "\e[1;36m";
        $normal = "\e[0m";
        $green = "\e[1;32m";
        $blue = "\e[1;34m";
    }

    $f_color = $malware->frequency >= $config->fn_freq_limit ? $red : $green;
    $r_color = $malware->per_unknown >= $config->fn_random_name_per ? $red : $green;
    $l_color = $malware->max_length >= $config->fn_line_limit ? $red : $green;

    echo "\n$red" . $malware->path . "$normal\n";
    echo "    Diff Size: $cyan" . $malware->file_size . "$normal bytes, File Size: $blue" . filesize($malware->path) . "$normal bytes";
    echo ", Modified: $blue" . date("Y-m-d H:i:s", filemtime($malware->path)) . $normal;
    echo ", Created: $cyan" . date("Y-m-d H:i:s", filectime($malware->path)) . "$normal\n";
    echo "    Note: \"$cyan" . $malware->note .  "$normal\"\n    frequency: ";
    echo $f_color . $malware->frequency . "$normal  unknown variable names: ";
    echo $r_color . $malware->per_unknown . "$normal%  longest line: ";
    echo $l_color . $malware->max_length . "$normal\n\e[1;30m";
}

if (!function_exists("readline")) {
    function readline($prompt) {
        echo $prompt;
        return trim(fgets(STDIN));
    }
}

function path_to_plugin(string $path) {
    $path = str_replace("\\", "/", $path);
    if (preg_match("/^.*\/wp-content\/plugins\/[^\/]+/", $path, $matches)) {
        return $matches[0];
    }
    return $path;
}

function repair(Malware $malware, bool $color, ScanConfig &$config) {
    $prompt = "Skip? (s)  Allow? (a)  Update? (u)  ";
    $valid = ["q", "d", "s", "a", "u"];
    if ($malware->known) {
        $prompt .= "Repair? (r)  ";
        $valid[] = "r";
    }
    $prompt .= "Quarantine? (q)  ";
    $prompt .= "Delete? (d)  ";
    $prompt .= "shOw? (o)  ";
    if (stristr($malware->path, "content/plugins/")) {
        $prompt .= "deactivate Plugin? (p)  ";
        $valid[] = "p";
    }
    if ($malware->known) {
        $prompt .= "  shOw? (o)";
    }

    $normal = $grey = "";
    if ($color) {
        $grey = "\e[1;30m";
        $normal = "\e[0m";
    }
    do {
        $input = readline($prompt);
        if ($input == "o") {
            echo "Diff with: " . $malware->url . "\n";
            echo $normal . $malware->pre_text .  $malware->content . "$grey\n";
        }
    } while (!in_array($input, $valid));

    if ($input == "s") {
        echo "Skipping...\n";
        return;
    }
    if ($input == "r") {
        $info = http2("GET", $malware->url);
        if ($info['success']) {
            file_put_contents($malware->path, $info['content']);
            echo $malware->path . " Repaired\n";
        } else {
            echo $malware->path . " Repair failed\n";
        }
    }
    if ($input == "a" || $input == "u") {
        $file_name = get_hidden_file("hashes.json");
        $ver_fn = '\BitFirePlugin\\version_from_path';
        $type_fn = "\BitFirePlugin\\file_type";
        $hash_file = hash_file3($malware->path, $type_fn, $ver_fn, Config::str("cms_root"));
        if (!file_exists($file_name)) { touch($file_name); }

        // load data and filter out this hash
        $file = FileData::new($file_name)->read()->un_json();
        $num1 = count($file->lines);

        $file->filter(function($x) use ($hash_file) { 
                $match = ($x["path"] == $hash_file->file_path && ($x["trim"] == $hash_file->crc_trim));
                return !$match;
            });
        $num2 = count($file->lines);

        // add the hash to the list
        $file->lines[] = [ "path" => $hash_file->file_path, "trim" => $hash_file->crc_trim ]; 
        $num3 = count($file->lines);
        debug("allow: %d -> %d -> %d", $num1, $num2, $num3);
        // all good, save the file
        Effect::new()->file(new FileMod($file_name, en_json($file->lines)))->run();
    }
    if ($input == "u") {
        //xdebug_break();
        $content = FileData::new($malware->path)->raw();
        // update the allow tokens
        //echo "[$content]\n";

        //$new_tokens = token_get_all($content);
        preg_match_all('/\\$[a-zA-Z_\x80-\xff][a-zA-Z0-9_\x80-\xff]*/mis', $content, $matches);
        $token_file = getcwd() . "/tokens.json";
        $src_file = file_exists($token_file) ? $token_file : WAF_ROOT . "cache/tokens2.json";
        //echo "src file: [$src_file]\n";
        $allow_tokens = FileData::new($src_file)->read()->un_json()->lines;
        //print_r($matches);
        foreach ($matches[0] as $match) {
            if (!empty($match[0])) {
                $allow_tokens[strtolower($match[0])] = 1;
            }
        }
        Effect::new()->file(new FileMod($token_file, en_json($allow_tokens, true)))->run();
        


        if ($malware->frequency > $config->fn_freq_limit) {
            $g_freq = array_fill(0, 128, []);
            $frequency = count_chars($content, 1);
            $lines = $frequency[10]??1;
        
            $freq_file = getcwd() . "/frequency.json";
            $src_file = file_exists($freq_file) ? $freq_file : WAF_ROOT . "cache/char_frequency.json";
            $freq = un_json(FileData::new($src_file)->raw());
            foreach ($frequency as $index => $count) {
                if ($index < 5) { continue; }
                if ($index <= 127) {
                    $per = $count / $lines;
                    if (isset($freq[$index])) {
                        if ($freq[$index]['u'] < $count) {
                            $freq[$index]['u'] = $count;
                        }
                        if ($freq[$index+128]['u'] < $per) {
                            $freq[$index+128]['u'] = $per;
                        }
                    }
                }
            }

            Effect::new()->file(new FileMod($freq_file, en_json($freq)))->run();
        }

        if ($malware->max_length > $config->line_limit) {
            $config->line_limit = floor($malware->max_length * 1.05);
        }
    }
    if ($input == "q") {
        $quarantine = preg_replace("/\.php$/", ".php.txt", $malware->path);
        if (rename($malware->path, $quarantine)) {
            echo "quarantined $malware->path to $quarantine\n";
        } else {
            echo "failed to rename, check permissions $malware->path\n";
        }
    }
    if ($input == "d") {
        if (unlink($malware->path)) {
            echo "deleted $malware->path\n";
        } else {
            echo "failed to delete, check permissions $malware->path\n";
        }
    }
    if ($input == "p") {
        $plugin_path = path_to_plugin($malware->path);
        if (file_exists($plugin_path)) {
            if (rename($plugin_path, $plugin_path . ".disabled")) {
                echo "disabled plugin $plugin_path\n";
                return;
            }
        }
        echo "unable to disable $plugin_path\n";
    }
}

define('DS', DIRECTORY_SEPARATOR);
if (!defined('BitFire\WAF_ROOT')) {
    define('BitFire\WAF_ROOT', __DIR__ . DIRECTORY_SEPARATOR);
    define('BitFire\WAF_SRC', \BitFire\WAF_ROOT . 'src' . DIRECTORY_SEPARATOR);
}


require_once __DIR__ . "/src/const.php";
require_once __DIR__ . "/src/bitfire.php";
require_once __DIR__ . "/src/util.php";
require_once __DIR__ . "/src/cms.php";
require_once __DIR__ . "/src/api.php";
require_once __DIR__ . "/src/diff.php";
require_once __DIR__ . "/src/includes.php";

$options = getopt("omhuUsip:b:aq:n:x:Q:N:l:L:");
// default config options
$config = new ScanConfig();
$config->freq_limit = 48;
$config->random_name_per = 40;
$config->fn_freq_limit = 20;
$config->fn_line_limit = 2048;
$config->fn_random_name_per = 20;

Config::set_value("debug_file", false);


if (isset($options['p'])) {
    $root = realpath($options['p']);
    if ($root == false) {
        $root = realpath(getcwd() . DIRECTORY_SEPARATOR . $options['p']);
    }
} else if (Config::str("cms_root")) {
    $root = Config::str("cms_root");
}
if (empty($options) || empty($root) || !isset($options['s'])) {
    usage($argv);
}
if (get_wordpress_version($root) == "1.0") {
    die("unable to locate wordpress install at [$root]\n");  
}
Config::set_value("cms_root", $root);

if (isset($options['l'])) { $config->fn_line_limit = intval($options['l']); }
if (isset($options['L'])) { $config->line_limit = intval($options['L']); }
if (isset($options['q'])) { $config->fn_freq_limit = intval($options['q']); }
if (isset($options['n'])) { $config->fn_random_name_per = intval($options['n']); }
if (isset($options['Q'])) { $config->freq_limit = intval($options['Q']); }
if (isset($options['N'])) { $config->random_name_per = intval($options['N']); }
if (!isset($options['u'])) { $config->unknown_core = false; }
if (!isset($options['U'])) { $config->unknown_plugin = false; }
else { $config->unknown_core = true; }
if (isset($options['a'])) {
    $display = 'print_r';
} else if (isset($options['o'])) {
    $display = function(Malware $m) { echo $m->path . "\n"; };
    $options['h'] = true;
} else {
    $display = partial_right('min_display', !isset($options['m']), $config);
}

//$display(new Malware());
//die("done");


$batch = 100;
if (isset($options['b'])) {
    $batch = intval($options['b']);
}

if (!$root || !file_exists($root)) {
    die("Path does not exist: $root");
}
$real_root = realpath($root);

config_display($config, $root, !isset($options['m']));
// default color terminal
if (!isset($options['m']) && !isset($options['a'])) { echo "\e[1;30m"; }

if (isset($options["s"])) {

    echo "indexing files...\n";
    $index_file = realpath("bitfire.index");
    $reg_ex = ($config->non_php == 1) ? NULL : "/.*\.php/";
    $ctx = fopen($index_file, "w+");
    $write_fn = partial('fwrite', $ctx);
    file_index($root, $reg_ex, $write_fn);
    fclose($ctx);


    $idx = 0;
    $done = false;
    while(!$done) {
        $list = scan_filesystem($index_file, $idx, $batch, $config);
        if ($list->complete) { $done = true; }
        if (count($list) > 0) { 
            foreach ($list as $malware) {
                $display($malware);
                if (isset($options['i'])) { 
                    repair($malware, !isset($options['m']), $config);
                }
            }
        }
        $idx += $batch;
        if (!isset($options['h'])) { echo  mt_rand(0,1) == 1 ? "X" : "x"; }
        //echo "$idx\n";
    }
}

// reset terminal
if (!isset($options['m'])) { echo "\e[0;m"; }
